function [f, x, y, z, info, xrec, yrec, zrec, trec, cpurec] = psapsr(A, epsln, radius, options)
% PSAPSR: Pseudospectral Abscissa or Pseudospectral Radius 
%
%  f = PSAPSR(A, epsln) normally returns
%    the epsln-pseudospectral ABSCISSA of the matrix A defined by
%      max {real(z): z is an eigenvalue of B with ||A-B|| <= epsln}
%    and is guaranteed to be a lower bound.  A can be dense or sparse.
%
%  f = PSASR(Afun, epsln) uses a function handle Afun instead of A. 
%    Afun must do matrix vector products with both A and A', as follows:
%    the function call Afun(p,t) returns A*p if t==0, A'*p if t==1.
%    For example, set Afun=@(p,t)amult(p,t) where the routine amult
%    returns A*p if t==0, A'*p if t==1. In all usages below, Afun
%    can be substituted for A.
%
%  f = PSAPSR(A, epsln, 1) normally returns
%    the epsln-pseudospectral RADIUS of the matrix A defined by
%      max {abs(z): z is an eigenvalue of B with ||A-B|| <= epsln}
%    and is guaranteed to be a lower bound.
%    A is either a matrix (usually sparse), or a function handle as
%    described above.
%
%  [f, x, y, z] = PSAPSR(A, epsln, radius) also returns column vectors x, y 
%   and scalar z satisfying: ||x||=||y||=1, 0 < y'x <= 1, 
%   epsln is the smallest singular value of A - zI, with corresponding 
%   right and left singular vectors x and -y, so (A - zI)x = -epsln y,
%   and x,y are right, left eigenvectors of B = A + epsln y*x', 
%   so Bx = zx,  y'B = zy'. 
%   When radius=0 (abscissa case, default), z is the rightmost point found 
%    in the epsln-pseudospectrum, and p = real(z).
%   When radius=1, (radius case), z is the largest modulus point found 
%    in the epsln-pseudospectrum, and p = abs(z).
% 
%  [f, x, y, z, info, xrec, yrec, zrec, trec, cpurec] = PSAPSR(A, epsln, radius) 
%  (abscissa case) also returns info:
%    0: normal return; 1: max iterations exceeded; 2: line search failure
%   as well as the record of all computed x, y and z iterates, the final 
%   values of t, and the CPU time/iteration.
%
%  [f, x, y, z] = PSAPSR(A, epsln, options) or 
%  [f, x, y, z] = PSAPSR(A, epsln, radius, options)
%   allows specification of the following options:
%  
%   options.maxit: max number of iterations (default 100)
%   options.ftol:   termination tolerance based on f (default 1e-8)
%   options.vectol: termination tolerance based on eigenvectors (default 0)
%   options.kstart: start iteration at kth rightmost or 
%     kth largest-modulus eigenvalue (default 1)
%   options.x0, options.y0: when A is a sparse matrix, or when A is
%     a function handle for computing matrix vector products A*v:
%     initial guesses for right and left eigenvectors for rightmost 
%     or largest-modulus eigenvalue of A (must be column vectors with
%     length the dimension of A).  These will be used for the first
%     call to "eigs".  At least one must be provided when  A is a 
%     function handle in order to specify the dimension of A, 
%     but it may be the zero vector if no estimate is known
%     (default: 0, which will be replaced by random starting vectors)
%   options.extra: number of additional eigenvalues to compute in "eigs", 
%     to help ensure finding the rightmost or largest-modulus eigenvalue.
%     May be good to set this to 1 if the imaginary part of the matrix
%     is tiny but not zero, since then there will be nearly conjugate 
%     eigenvalues, and want to ensure the eigenvalues found for A and its 
%     transpose are nearly the same, not nearly conjugates. 
%     Ignored if A is dense as then all eigenvalues are computed.  
%     (default: 0, for efficiency)
%   options.monotone: 1 for monotonically increasing iterates (default)
%                        (Algorithms PSA1 and PSR1 of paper)
%                     0 for basic iteration: no guarantees
%                        (Algorithms PSA0 and PSR0 of paper)
%   options.prtlevel: print level 
%     0: no output; 1: minimal output (default); 2: verbose output
%
%   Primarily intended for large sparse A or other matrix A for which
%   matrix-vector products A*v, A'*v are cheap.  Can also be used for 
%   small dense A, but in that case a more reliable code can be found at
%   http://www.cs.nyu.edu/mengi/robuststability.html
%
%   Reference: Nicola Guglielmi and Michael Overton, Fast Algorithms 
%    for the Approximation of the Pseudospectral Abscissa and 
%    Pseudospectral Radius of a Matrix, submitted to SIAM J Mat Anal Appl
%   Send comments/bug reports to Michael Overton, overton@cs.nyu.edu,
%   with a subject header containing the string "psapsr".
%
%
%   PSAPSR Version 1.4, 2016, see GPL license info below.
%   psapsr.m introduced in PSAPSR Version 1.0.
%
%##########################################################################
%#  PSAPSR: PseudoSpectral Abscissa and PseudoSpectral Radius             #
%#  Copyright (C) 2016 Nicola Guglielmi, Michael Overton, Tim Mitchell    #
%#                                                                        #
%#  This file is part of PSAPSR.                                          #
%#                                                                        #
%#  PSAPSR is free software: you can redistribute it and/or modify it     #
%#  under the terms of the GNU General Public License as published by     #
%#  the Free Software Foundation, either version 3 of the License, or     #
%#  (at your option) any later version.                                   #
%#                                                                        #
%#  PSAPSR is distributed in the hope that it will be useful, but         #
%#  WITHOUT ANY WARRANTY; without even the implied warranty of            #
%#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
%#  GNU General Public License for more details.                          #
%#                                                                        #
%#  You should have received a copy of the GNU General Public License     #
%#  along with this program.  If not, see <http://www.gnu.org/licenses/>. #
%##########################################################################

% Tim Mitchell
% handle variable input args A,epsln,radius,options 
% fixes minor input argument handling present in earlier versions
switch nargin
    case 1
        error('both A and epsln are required input parameters')
    case 2
        radius = false; % 2 args: default objective is spectral abscissa
        options = []; % no options specified
    case 3  % 3 args, determine if radius or options given
        if isstruct(radius)
            options = radius;
            radius = false;  % radius not given, default to abscissa
        elseif islogical(radius) || radius == 0 || radius == 1
            options = []; % no options given
        else
            error('3rd argument must either be radius (boolean) or options (struct)')
        end
    otherwise
        % for 4 (or more) input arguments, additonal args beyond 4th are 
        % ignored. Confirm correct types for radius and options
        if ~(islogical(radius) || radius == 0 || radius == 1) || ~isstruct(options)
            error('3rd argument must be radius (boolean) and 4th argument must be options (struct)');
        end
end
if isa(A, 'double')
    n = size(A,1);
    if size(A,2) ~= n
        error('the matrix A must be square')
    end
    if ~issparse(A)
        Atype = 1; % A is dense
    else
        Atype = 2; % A is sparse
    end
elseif isa(A, 'function_handle')
    Atype = 3; % A is a function handle for computing products A*w
else
    error('A must either be a matrix or a function handle')
end
if ~isa(epsln, 'double') || epsln < 0 || imag(epsln) ~= 0
    error('epsln must be nonnegative')
end
if isfield(options, 'maxit')
    maxit = options.maxit;
else
    maxit = 100;
end
if isfield(options, 'ftol')
    ftol = options.ftol;
else
    ftol = 1e-8; % absolute and relative tolerance on f 
end
if isfield(options, 'vectol')
    vectol = options.ftol;
else
    vectol = 0; % 0 means turned off
end
if isfield(options, 'kstart')
    kstart = options.kstart;
else
    kstart = 1;
end
if isfield(options, 'extra')
    extra = options.extra;
else
    extra = 0;
end
if isfield(options, 'monotone')
    monotone = options.monotone;
else
    monotone = 1; % guarantees monotonicity
end
if isfield(options, 'x0')
    x0 = options.x0;
    if size(x0,2) ~= 1
        error('options.x0 must be a column vector')
    end
    if exist('n','var')
        if length(x0) ~= n
            error('length of options.x0 is not compatible with size of A')
        end
    else
        n = length(x0);
    end
else
    if exist('n','var')
        x0 = zeros(n, 1); % if not, hold off until we see if options.y0 specified
    end
end
if isfield(options, 'y0')
    y0 = options.y0;
    if size(y0,2) ~= 1
        error('options.y0 must be a column vector')
    end
    if exist('n','var')
        if length(y0) ~= n
            error('length of options.y0 is not compatible with size of A or length of options.x0')
        end
    else
        n = length(y0);
        if ~exist('x0','var')
            x0 = zeros(n,1); % now that we know what n is
        end
    end
else
    if exist('n','var')
        y0 = zeros(n, 1);
    else
        error('options.x0 or options.y0 must be provided when A is a function handle')
    end
end
if isfield(options, 'prtlevel')
    prtlevel = options.prtlevel;
else
    prtlevel = 1;
end
if prtlevel > 0
    if ~radius
        fprintf('Computing pseudospectral abscissa: n = %d, epsln = %g, Atype = %d, monotone = %d\n', n, epsln, Atype, monotone)
    else
        fprintf('Computing pseudospectral radius: n = %d, epsln = %g, Atype = %d, monotone = %d\n', n, epsln, Atype, monotone)
    end
end
if extra > n-3
    extra = max([0, n-3]); % otherwise get warning from eigs
end
%
% compute right and left eigenvectors x0, y0 
% for kstart-th largest real part or largest modulus eigenvalue z0 of A,
% normalized so ||x0||=||y0||=1 and y'x is real and positive

% BUG FIX 11/7/13: Tim Mitchell - warm starting with options.x0 and 
% options.y0 was broken because: 
%  a) scalefac was erroneously set to 0 so it was ignored!
%  b) the x,y swap procedure on the first step needs to be skipped when 
%     warm starting
%  c) no check was done to ensure options.x0 and options.y0 have norm 1
scalefac = epsln;
cold_start = true;
if any(x0) && any(y0)
    % only warm start when outer product of x0 and y0 is non-zero 
    cold_start = false; 
    % need to make sure options.x0 and options.y0 have norm 1 so that
    % epsln*y0*x0' is a matrix perturbation of norm epsln
    x0 = x0 / norm(x0);
    y0 = y0 / norm(y0);
end

% we also need spectral abscissa or radius to check if first step is any good
cputime0 = cputime; % cputime is built in function
% BUG FIX 11/22/16: Tim Mitchell
% abs(y'*x) is needed for the eigenvector swap that may happen on the first
% perturbation from the matrix A (does not apply if PSAPSR is warm started)
[z, x, y, f, absyhx] = geteigvecs(A, Atype, scalefac, x0, y0, radius, kstart, extra);

% BUG FIX 11/7/13: Tim Mitchell - warm starting with options.x0, options.y0
% need to set xold and yold for monotone line search between eigenvalue z
% of A + scalefac*y0*x0' and subsequent znew of A + scalefac*y*x'
xold = x;
yold = y;

% from now on we will not use kstart: we will only want the largest
% eigenvalue wrt to real part or modulus
t = 0; % there is no t at initial iterate
for iter = 1:maxit
    if prtlevel > 1
        fprintf('iter %d, f = %g   \n', iter, f)
    end
    if nargout > 4 % don't preallocate since don't expect many iterations
        zrec(iter) = z; 
        xrec(:,iter) = x;
        yrec(:,iter) = y;
        cputime1 = cputime;  
        trec(iter) = t;
        cpurec(iter) = cputime1 - cputime0;
        cputime0 = cputime1;
    end
    if epsln == 0 % cannot move before loop in case nargout > 4
        % pseudospectrum is just the spectrum
        info = 0;
        return
    end
    % for iter = 1, choose u, v to maximizer derivative of spectral abscissa 
    % or radius of A + t u*v', and for iter > 1, choose u, v to maximize derivative 
    % of spectral absicssa or spectral radius of B + t(u*v' - yold*xold')
    % where B = A + yold*xold'
    % in both cases, this means setting u and v to the new y, x 
    scalefac = epsln;
    [znew,xnew,ynew,fnew] = geteigvecs(A, Atype, scalefac, x, y, radius, 1, extra); 
    % termination check based on the eigenvector
    if iter > 1 && vectol > 0 
        if Atype == 1 
            yxdif = max(max(abs(ynew*xnew' - y*x')));
        else
            xrat = xnew./x;
            yrat = ynew./y;
            xrat = xrat/xrat(1);
            yrat = yrat/yrat(1);
            one = ones(n,1);
            yxdif = norm(xrat - one) + norm(yrat - one);
        end
        if yxdif < vectol
            if prtlevel > 0
                fprintf('terminating: yxdif = %g at iteration %d\n', yxdif, iter)
            end
            info = 0; % normal termination
            break % out of iteration
        end
    end
    % termination check based on the function value
    if iter > 1 && abs(fnew - f) < ftol*max([1 abs(f)])
        if prtlevel > 0
            fprintf('terminating: fnew - f = %g at iteration %d\n', fnew - f, iter)
        end
        info = 0; % normal termination
        break % out of iteration
    end
    % special check for first iteration, regardless of options.monotone
    % BUG FIX 11/7/13: Tim Mitchell - swapping trick is only applicable 
    % for eigenvectors and we don't know a priori if x0 and y0 specified 
    % by the user are eigenvectors or not and even if they are, we don't 
    % know what specific matrix (A + perturbation) they are eigevectors of. 
    % So only check and possibly do swap if we are cold starting without 
    % userset x0 and y0; otherwise skip and proceed to monotone check if 
    % it's enabled.
    if iter == 1 && cold_start
        % BUG FIX 11/22/16: Tim Mitchell
        % This previously used ytx = y'*x which made the conditional
        % meaningless since ytx could be complex for the radius case!
        % geteigvecs now returns absyhx = abs(y'*x) to be used below.
        if fnew < f + epsln*absyhx
            % initial step is no good: swap x and y so eigenvalue variation is  linear
            % then the abscissa/radius increases by at least epsln on the first step
            if prtlevel > 0
                fprintf('insufficient increase at first step: swapping yx* for xy*\n')
            end
            % instead, y and x are swapped
            xtemp = x; x = y; y = xtemp; % this also ensures xold, yold get set correctly below
            [znew,xnew,ynew,fnew] = geteigvecs(A, Atype, scalefac, x, y, radius, 1, extra); 
            if fnew < f + epsln*absyhx
                error('psapsr: something went wrong with the swap')
            end
        end
    elseif monotone % for iter > 1, bisect t if necessary
    % the next part was incorrectly stated in Algorithm PSA1/PSR1 in the
    % paper and was wrongly implemented in version 1.01 of the code
    % fnew is the spectral function of A + eps y*x'
        t = 1;
        xtriedfirst = x;  % the first attempt at x_k, y_k
        ytriedfirst = y;
        while fnew < f
            t = t/2;
            if prtlevel > 0
                fprintf('f decreased by %g so reduce t to %g\n', f-fnew, t)
            end
            if t < 0.001
                if prtlevel > 0
                    fprintf('terminating since t is small: may have reached limits of precision\n')
                end
                info = 2;
                return
            end
            % xold, yold are x_{k-1}, y_{k-1} defining B_k = A + eps y_{k-1}x_{k-1}'
            % for which we did get an increase in f from previous iteration
            x = t*xtriedfirst + (1-t)*xold;  % this was wrong in version 1.01
            x = x/norm(x);
            y = t*ytriedfirst + (1-t)*yold;
            y = y/norm(y);
            [znew,xnew,ynew,fnew] = geteigvecs(A, Atype, epsln, x, y, radius, 1, extra);
        end 
    end  
    % check monotonicity
    if fnew < f
        if iter == 1
            error('decrease in f should not happen since swap should have taken place, see above')
        elseif monotone
            warning('f decreased, but there should be a monotone increase in f')
            keyboard
        end
        if prtlevel > 0 % could happen otherwise
            fprintf('f decreased by %g at iteration %d\n', f-fnew, iter)
        end
    end
    % at next step we will computer eigenvalues of B = A + epsln*ynew*xnew'
    xold = x; % save for use if options.monotone is 1
    yold = y;
    x = xnew;
    y = ynew;
    z = znew;
    f = fnew;
end
if iter == maxit
    info = 1;
    if prtlevel > 0
        fprintf('reached %d iterations, stop, f = %g\n',maxit, fnew)
    end
end
if nargout > 4
    zrec = [zrec znew];
    xrec = [xrec xnew];
    yrec = [yrec ynew];
    cpurec = [cpurec cputime - cputime0];
end
if ~radius
    f = real(znew); % not z, if tolerance was satisfied
else
    f = abs(znew);
end